use std::{
    path::Path,
    sync::{Arc, LazyLock},
};

use protocol::protocol::{Signal, Speaker};
use s2n_quic::{
    Connection, Server,
    stream::{ReceiveStream, SendStream},
};
use tokio::{
    io::{AsyncReadExt, AsyncWriteExt},
    sync::{RwLock, broadcast},
};

use crate::ServerConfig;

const NEW_SPEAKER_LENGTH: usize = u8::MAX as usize;
const AUDIO_BUFFER_LENGTH: usize = 1024 * 16 * 16;

#[derive(Debug, Clone, Copy)]
enum SpeakerAction {
    Join,
    Left,
}

#[derive(Debug, Clone)]
struct SpeakerWithData {
    speaker: Speaker,
    speaker_action_sender: broadcast::Sender<(SpeakerWithData, SpeakerAction)>,
    audio_data_sender: broadcast::Sender<f32>,
}

impl SpeakerWithData {
    async fn new(
        speaker: Speaker,
    ) -> (
        broadcast::Receiver<(SpeakerWithData, SpeakerAction)>,
        broadcast::Sender<f32>,
    ) {
        let speaker_action_channel = broadcast::channel(NEW_SPEAKER_LENGTH);
        let audio_data_sender = broadcast::channel(AUDIO_BUFFER_LENGTH).0;

        let speaker_with_data = Self {
            speaker,
            speaker_action_sender: speaker_action_channel.0,
            audio_data_sender: audio_data_sender.clone(),
        };

        let mut online_speakers = ONLINE_SPEAKERS.write().await;
        for online_speaker in online_speakers.iter() {
            let _ = speaker_with_data
                .speaker_action_sender
                .send((online_speaker.clone(), SpeakerAction::Join));
            let _ = online_speaker
                .speaker_action_sender
                .send((speaker_with_data.clone(), SpeakerAction::Join));
        }

        online_speakers.push(speaker_with_data);
        online_speakers.sort_by_key(|speaker| speaker.speaker.get_id());

        drop(online_speakers);

        (speaker_action_channel.1, audio_data_sender)
    }

    async fn remove(speaker_id: u8) {
        let mut online_speakers = ONLINE_SPEAKERS.write().await;

        let speaker_index =
            online_speakers.binary_search_by_key(&speaker_id, |speaker| speaker.speaker.get_id());

        match speaker_index {
            Ok(speaker_index) => {
                let speaker = online_speakers.remove(speaker_index);
                for online_speaker in online_speakers.iter() {
                    let _ = online_speaker
                        .speaker_action_sender
                        .send((speaker.clone(), SpeakerAction::Left));
                }
            }
            Err(_) => return,
        }
    }
}

static ONLINE_SPEAKERS: LazyLock<RwLock<Vec<SpeakerWithData>>> = LazyLock::new(|| vec![].into());

pub async fn start(server_config: ServerConfig) {
    let mut server = Server::builder()
        .with_io(server_config.server_address.as_str())
        .unwrap()
        .with_tls((
            Path::new(&server_config.certificate_path),
            Path::new(&server_config.certificate_key_path),
        ))
        .unwrap()
        .start()
        .unwrap();

    let mut speaker_id = 0;
    while let Some(connection) = server.accept().await {
        let speaker = Speaker::new(speaker_id);
        tokio::spawn(handle_client(speaker, connection));

        match speaker_id.checked_add(1) {
            Some(next_speaker_id) => speaker_id = next_speaker_id,
            None => {
                println!("Warning: Room Is Full");
                return;
            }
        }
    }
}

async fn handle_client(speaker: Speaker, mut connection: Connection) {
    println!(
        "Server Name = {}",
        connection.server_name().unwrap().unwrap().to_string()
    );

    println!(
        "Client Address = {}",
        connection.remote_addr().unwrap().to_string()
    );

    let stream = connection
        .accept_bidirectional_stream()
        .await
        .unwrap()
        .unwrap();
    let (receive_stream, send_stream) = stream.split();

    let speaker_id = speaker.get_id();
    let (speaker_action_receiver, audio_data_sender) = SpeakerWithData::new(speaker).await;

    tokio::spawn(receive_audio_data(
        receive_stream,
        speaker_id,
        audio_data_sender,
    ));
    tokio::spawn(send_audio_data(
        send_stream,
        speaker_id,
        speaker_action_receiver,
    ));
}

async fn send_audio_data(
    send_stream: SendStream,
    speaker_id: u8,
    mut speaker_action_receiver: broadcast::Receiver<(SpeakerWithData, SpeakerAction)>,
) {
    let send_stream = Arc::new(RwLock::new(send_stream));
    while let Ok((speaker_with_data, speaker_action)) = speaker_action_receiver.recv().await {
        match speaker_action {
            SpeakerAction::Join => {
                let send_stream = send_stream.clone();
                tokio::spawn(async move {
                    let mut audio_data_receiver = speaker_with_data.audio_data_sender.subscribe();

                    while let Ok(audio_datum) = audio_data_receiver.recv().await {
                        let data = Signal::pack_audio_datum(speaker_with_data.speaker, audio_datum);
                        if let Err(err_val) = send_stream.write().await.write_all(&data).await {
                            eprintln!("Error: Send Audio Data | Remote | {}", err_val);
                            SpeakerWithData::remove(speaker_id).await;
                        };
                    }
                });
            }
            SpeakerAction::Left => {
                let data = Signal::pack_speaker_left(speaker_with_data.speaker);
                if let Err(err_val) = send_stream.write().await.write_all(&data).await {
                    eprintln!("Error: Send Speaker Left | Remote | {}", err_val);
                    SpeakerWithData::remove(speaker_id).await;
                }
            }
        }
    }
}

async fn receive_audio_data(
    mut receive_stream: ReceiveStream,
    speaker_id: u8,
    audio_data_sender: broadcast::Sender<f32>,
) {
    loop {
        match receive_stream.read_f32().await {
            Ok(received_data) => {
                let _ = audio_data_sender.send(received_data);
            }
            Err(err_val) => {
                eprintln!("Error: Receive Audio Data | Remote | {}", err_val);
                SpeakerWithData::remove(speaker_id).await;
            }
        }
    }
}
